--!parent
--!encoding=utf-8

--****************************************************************************
-- @file      Rich Text Editor - main class
-- @author    Valentin Schmidt
-- @version   0.4
-- @todo
-- - textStyles (bullet lists etc.) not implemented yet (-> ComboStyle)
-- - save/restore window state and geometry across sessions
-- - localization 
-- - maybe add basic syntax highlighting?
-- - maybe add basic hex view/edit support?
-- - what about multi-file support and tabs?
-- - automatic online updates?
--****************************************************************************

global $

property pAppName

-- widgets
property pMainWindow
property pTextEdit
property pMenuBar
property pComboFont
property pComboSize

-- actions
-- file
property pActionSave
-- edit
property pActionUndo
property pActionRedo
--
property pActionCut
property pActionCopy
property pActionPaste
-- format
property pActionTextBold
property pActionTextItalic
property pActionTextUnderline
--
property pActionAlignLeft
property pActionAlignHCenter
property pActionAlignRight
property pActionAlignJustify
--
property pActionColor

-- other qt objects
property pClipboard
property pPrinter
property pDefaultFont

-- other vars
property pTextColor
property pFileName
property pIconPath
property pTextTypes

----------------------------------------
-- @constructor
----------------------------------------
on new(me)

  me.pAppName = "EditorXX"
	
  -- SET DEFAULTS
  me.pDefaultFont = $.Qt.Font.new()
  me.pDefaultFont.setFamily("Courier New")
  me.pDefaultFont.setPointSize(10.0) -- Arial
  me.pDefaultFont.setBold(FALSE)
  me.pDefaultFont.setItalic(FALSE)
  me.pDefaultFont.setUnderline(FALSE)

	me.pTextTypes = ["txt","ls","lsw","php","js","jsw","py","pyw","ini"]
	
  me.pIconPath = $.@("images/"&$.os&"/")

  -- CREATE MAIN WINDOW
  me.pMainWindow = $.Qt.MainWindow.new()
  me.pMainWindow.connect($.Qt.Event.Close, me, #fileQuit)

  me.pMainWindow.setWindowIcon($.@("images/logo.png"))
  me.pMainWindow.setToolButtonStyle($.Qt.ToolButtonFollowStyle)

  -- ADD DROP SUPPORT	
	me.pMainWindow.setAcceptDrops(TRUE)

  -- CREATE TEXTEDIT
  me.pTextEdit = $.Qt.TextEdit.new(me.pMainWindow)
  me.pMainWindow.setCentralWidget(me.pTextEdit)

  -- CREATE MENUBAR
  me.pMenuBar = $.Qt.MenuBar.new(me.pMainWindow)
  me.pMainWindow.setMenuBar(me.pMenuBar)

	me.setupFileActions()
	me.setupEditActions()
	me.setupFormatActions()
	me.setupHelpActions()

  -- CREATE STATUSBAR
  statusBar = $.Qt.StatusBar.new(pMainWindow)
  me.pMainWindow.setStatusBar(statusBar)

  -- WATCH SYSTEM CLIPBOARD FOR CHANGES
  me.pClipboard = $.Qt.Clipboard.new(me.pMainWindow)
  me.pClipboard.connect(#dataChanged, me, #clipboardDataChanged)

  -- CREATE A PRINTER OBJECT
  me.pPrinter = $.Qt.Printer.new(0, $.Qt.QPrinter.HighResolution)

  -- SETUP TEXTEDIT SIGNALS
  me.pTextEdit.connect(#cursorPositionChanged, me, #cursorPositionChanged)
  me.pTextEdit.connect(#currentCharFormatChanged, me, #currentCharFormatChanged)
  me.pTextEdit.connect(#copyAvailable, me, #copyAvailable)
	me.pTextEdit.connect(#modificationChanged, me, #modificationChanged) -- ???

  -- LOAD FILE?

  -- A) args parsed by lingo
	if $.args.count>1 then me.load($.args[2])
	else me.fileNew()
	
	--me.pActionSave.setEnabled(FALSE)

	--me.pTextEdit.connect(#modificationChanged, me, #modificationChanged) -- ???

  -- B) args parsed by Qt ($.includes EXE as first arg)
  -- app = $.Qt.Application.new()
  -- args = app.arguments()
  -- if args.count>2 then me.load(args[3])
  -- else me.fileNew()

  -- set cwd?
  -- app = $.Qt.Application.new()
  -- cwd = app.applicationDirPath()
  cwd = $.path -- LSW script dir
  $.Qt.Dir.new(#Dir).setCurrent(cwd)

	me.pTextEdit.connect(#undoAvailable, me, #undoAvailable)
	me.pTextEdit.connect(#redoAvailable, me, #redoAvailable)
	
  -- RESIZE AND SHOW MAIN WINDOW
	settings = $.Qt.Settings.new("Fluxus","Editor",1) -- 0=Reg.(=def.), 1=INI
	
	state = value(settings.value("mainwindow/state"))
	if state=2 then
		me.pMainWindow.resize(680, 680)
		me.pMainWindow.setWindowState(state)
	else
	  rect = value(settings.value("mainwindow/rect"))
	  if ilk(rect)=#rect then me.pMainWindow.setGeometry(rect)
	  else me.pMainWindow.resize(680, 680)
  end if
    
  me.pMainWindow.show()
    
  return me
end

--------------------------------------
--
--------------------------------------
on setupFileActions(me)
	
  toolBar = $.Qt.ToolBar.new(me.pMainWindow)
  toolBar.setWindowTitle("File Actions")
  
  me.pMainWindow.addToolBar(toolBar)

  menu = me.pMenuBar.addMenu("&File")

  action = $.Qt.Action.new(menu)
  action.setText("&New")
  action.setIcon(me.pIconPath & "filenew.png")
  action.connect(#triggered, me, #fileNew)
  action.setShortcut("Ctrl+N")
  
  toolBar.addAction(action)
  menu.addAction(action)

  action = $.Qt.Action.new(menu)
  action.setText("&Open")
  action.setIcon(me.pIconPath & "fileopen.png")
  action.connect(#triggered, me, #fileOpen)
  action.setShortcut("Ctrl+O")
  toolBar.addAction(action)
  menu.addAction(action)

  menu.addSeparator()

  action = $.Qt.Action.new(menu)
  action.setText("&Save")
  action.setIcon(me.pIconPath & "filesave.png")
	action.connect(#triggered, me, #fileSave)
  action.setShortcut("Ctrl+S")
  toolBar.addAction(action)
  menu.addAction(action)
  me.pActionSave = action

  action = $.Qt.Action.new(menu)
  action.setText("Save &As...")
  action.setIcon(me.pIconPath & "filesave.png")
	action.connect(#triggered, me, #fileSaveAs)
  action.setShortcut("Ctrl+Shift+S")
  menu.addAction(action)

  menu.addSeparator()

  action = $.Qt.Action.new(menu)
  action.setText("&Export ODF...")
  action.setIcon(me.pIconPath & "exportodf.png")
	action.connect(#triggered, me, #fileExportODF)
  toolBar.addAction(action)
  menu.addAction(action)

  action = $.Qt.Action.new(menu)
  action.setText("E&xport PDF...")
  action.setIcon(me.pIconPath & "exportpdf.png")
	action.connect(#triggered, me, #filePrintPdf)
  toolBar.addAction(action)
  menu.addAction(action)

  menu.addSeparator()

  action = $.Qt.Action.new(menu)
  action.setText("Print Preview...")
	action.connect(#triggered, me, #filePrintPreview)
  action.setShortcut("Ctrl+Shift+P")
  menu.addAction(action)

  action = $.Qt.Action.new(menu)
  action.setText("&Print...")
  action.setIcon(me.pIconPath & "fileprint.png")
	action.connect(#triggered, me, #filePrint)
  action.setShortcut("Ctrl+P")
  toolBar.addAction(action)
  menu.addAction(action)

  menu.addSeparator()

  action = $.Qt.Action.new(menu)
  action.setText("&Quit")
	action.connect(#triggered, me, #fileQuit)
  menu.addAction(action)
end

--------------------------------------
--
--------------------------------------
on setupEditActions(me)

  toolBar = $.Qt.ToolBar.new(me.pMainWindow)
  toolBar.setWindowTitle("Edit Actions")
  me.pMainWindow.addToolBar(toolBar)
  toolBar.setVisible(FALSE)

  menu = pMenuBar.addMenu("&Edit")

  action = $.Qt.Action.new(menu)
  action.setText("&Undo")
  action.setIcon(me.pIconPath & "editundo.png")
	action.connect(#triggered, me, #undo)
  action.setShortcut("Ctrl+Z")
  toolBar.addAction(action)
  menu.addAction(action)
  action.setEnabled(FALSE)
  me.pActionUndo = action

  action = $.Qt.Action.new(menu)
  action.setText("&Redo")
  action.setIcon(me.pIconPath & "editredo.png")
  action.connect(#triggered, me, #redo)
  action.setShortcut("Ctrl+Y")
  toolBar.addAction(action)
  menu.addAction(action)
  action.setEnabled(FALSE)
  me.pActionRedo = action

  menu.addSeparator()

  action = $.Qt.Action.new(menu)
  action.setText("Cu&t")
  action.setIcon(me.pIconPath & "editcut.png")
	action.connect(#triggered, me, #cut)
  action.setShortcut("Ctrl+X")
  toolBar.addAction(action)
  menu.addAction(action)
  action.setEnabled(FALSE)
  me.pActionCut = action

  action = $.Qt.Action.new(menu)
  action.setText("&Copy")
  action.setIcon(me.pIconPath & "editcopy.png")
	action.connect(#triggered, me, #copy)
  action.setShortcut("Ctrl+C")
  toolBar.addAction(action)
  menu.addAction(action)
  action.setEnabled(FALSE)
  me.pActionCopy = action

  action = $.Qt.Action.new(menu)
  action.setText("&Paste")
  action.setIcon(me.pIconPath & "editpaste.png")
	action.connect(#triggered, me, #paste)
  action.setShortcut("Ctrl+V")
  toolBar.addAction(action)
  menu.addAction(action)
  me.pActionPaste = action

end

--------------------------------------
--
--------------------------------------
on setupFormatActions(me)

  toolBar = $.Qt.ToolBar.new(me.pMainWindow)
  toolBar.setWindowTitle("Format Actions")
  me.pMainWindow.addToolBar(toolBar)

  menu = me.pMenuBar.addMenu("&Format")

  -- BOLD
  action = $.Qt.Action.new(menu)
  action.setText("&Bold")
  action.setIcon(me.pIconPath & "textbold.png")
	action.connect(#triggered, me, #textBold)
  action.setShortcut("Ctrl+B")
  action.setCheckable(TRUE)
  font = action.getFont()
  font.setBold(TRUE)
  action.setFont(font)
  toolBar.addAction(action)
  menu.addAction(action)
  me.pActionTextBold = action

  -- ITALIC
  action = $.Qt.Action.new(menu)
  action.setText("&Italic")
  action.setIcon(me.pIconPath & "textitalic.png")
 	action.connect(#triggered, me, #textItalic)
  action.setShortcut("Ctrl+I")
  action.setCheckable(TRUE)
  font = action.getFont()
  font.setItalic(TRUE)
  action.setFont(font)
  toolBar.addAction(action)
  menu.addAction(action)
  me.pActionTextItalic = action

  -- UNDERLINE
  action = $.Qt.Action.new(menu)
  action.setText("&Underline")
  action.setIcon(me.pIconPath & "textunder.png")
	action.connect(#triggered, me, #textUnderline)
  action.setShortcut("Ctrl+U")
  action.setCheckable(TRUE)
  font = action.getFont()
  font.setUnderline(TRUE)
  action.setFont(font)
  toolBar.addAction(action)
  menu.addAction(action)
  me.pActionTextUnderline = action

  menu.addSeparator()
  toolBar.addSeparator()

  actionGroup = $.Qt.ActionGroup.new(menu)
	actionGroup.connect(#triggered, me, #textAlign)

  action = $.Qt.Action.new(menu)
  action.setText("&Left")
  action.setIcon(me.pIconPath & "textleft.png")
  action.setShortcut("Ctrl+L")
  action.setCheckable(TRUE)
  toolBar.addAction(action)
  menu.addAction(action)
  actionGroup.addAction(action)
  me.pActionAlignLeft = action

  action = $.Qt.Action.new(menu)
  action.setText("C&enter")
  action.setIcon(me.pIconPath & "textcenter.png")
  action.setShortcut("Ctrl+E")
  action.setCheckable(TRUE)
  toolBar.addAction(action)
  menu.addAction(action)
  actionGroup.addAction(action)
  me.pActionAlignHCenter = action

  action = $.Qt.Action.new(menu)
  action.setText("&Right")
  action.setIcon(me.pIconPath & "textright.png")
  action.setShortcut("Ctrl+R")
  action.setCheckable(TRUE)
  toolBar.addAction(action)
  menu.addAction(action)
  actionGroup.addAction(action)
  me.pActionAlignRight = action

  action = $.Qt.Action.new(menu)
  action.setText("&Justify")
  action.setIcon(me.pIconPath & "textjustify.png")
  action.setShortcut("Ctrl+J")
  action.setCheckable(TRUE)
  toolBar.addAction(action)
  menu.addAction(action)
  actionGroup.addAction(action)
  me.pActionAlignJustify = action

  menu.addSeparator()
  toolBar.addSeparator()

  -- COLOR
  img = image(16,16,32)
  img.fill(img.rect, color(0,0,0))
  me.pActionColor = $.Qt.Action.new(menu)
  me.pActionColor.setText("&Color...")
  me.pActionColor.setIcon(img)
	me.pActionColor.connect(#triggered, me, #textColor)
  toolBar.addAction(me.pActionColor)
  menu.addAction(me.pActionColor)

  -- Font Actions Toolbar

  toolBar = $.Qt.ToolBar.new(pMainWindow)
  toolBar.setAllowedAreas(bitOr($.Qt.TopToolBarArea, $.Qt.BottomToolBarArea))
  toolBar.setWindowTitle("Font Actions")
  --pMainWindow.addToolBarBreak($.Qt.TopToolBarArea)
  pMainWindow.addToolBar(toolBar)

  -- ComboStyle -------------------- TODO

--  comboStyle = $.Qt.ComboBox.new(toolBar)
--  toolBar.addWidget(comboStyle)
--
--  comboStyle.addItem("Standard")
--  comboStyle.addItem("Bullet List (Disc)")
--  comboStyle.addItem("Bullet List (Circle)")
--  comboStyle.addItem("Bullet List (Square)")
--  comboStyle.addItem("Ordered List (Decimal)")
--  comboStyle.addItem("Ordered List (Alpha lower)")
--  comboStyle.addItem("Ordered List (Alpha upper)")
--  comboStyle.addItem("Ordered List (Roman lower)")
--  comboStyle.addItem("Ordered List (Roman upper)")
--
--  comboStyle.connect(#activated, me, #textStyle)
		
  -- ComboFont --------------------

  me.pComboFont = $.Qt.FontComboBox.new(toolBar)
  toolBar.addWidget(me.pComboFont)

	me.pComboFont.connect(#activated, me, #textFamily)

  -- ComboSize --------------------

  me.pComboSize = $.Qt.ComboBox.new(toolBar)
  toolBar.addWidget(me.pComboSize)
  me.pComboSize.setObjectName("comboSize")
  me.pComboSize.setEditable(TRUE)
  standardSizes = [6,7,8,9,10,12,14,16,18,20,22,24,26,28,36,48,72]
  repeat with s in standardSizes
    me.pComboSize.addItem(string(s))
  end repeat

  me.pComboSize.connect(#activated, me, #textSize)

end

--------------------------------------
--
--------------------------------------
on setupHelpActions(me)

  menuHelp = me.pMenuBar.addMenu("Help")

  action = $.Qt.Action.new(menuHelp)
  action.setText("Show Console")
  action.connect(#triggered, me, #helpConsole)
  menuHelp.addAction(action)
  
  action = $.Qt.Action.new(menuHelp)
  action.setText("Edit me!")
  action.connect(#triggered, me, #helpEdit)
  menuHelp.addAction(action)
  
  menuHelp.addSeparator()
  
  action = $.Qt.Action.new(menuHelp)
  action.setText("About &"&me.pAppName)
  action.connect(#triggered, me, #helpAbout)
  menuHelp.addAction(action)

  action = $.Qt.Action.new(menuHelp)
  action.setText("About &Qt")
  action.connect(#triggered, me, #helpAboutQt)
  menuHelp.addAction(action)
 end

--------------------------------------
--
--------------------------------------
on resetDefaults(me)

  pTextEdit.clear()

  -- reset font
  me.pTextEdit.setFont(me.pDefaultFont)
  me.pTextEdit.setCurrentFont(me.pDefaultFont)

  -- set tabStop width
  metrics = me.pTextEdit.getFontMetrics()
  w = metrics.width(" ")
  me.pTextEdit.setTabStopWidth(2*w)

  -- set align
  me.pTextEdit.setAlignment(bitOr($.Qt.AlignLeft,$.Qt.AlignAbsolute))
  me.pActionAlignLeft.setChecked(TRUE)

  -- set color
  me.pTextEdit.setTextColor(color(0,0,0))
end

--------------------------------------
--
--------------------------------------
on load(me, fn)
	
  data = $.file.getBytes(fn)
  if data=false then return FALSE

  me.resetDefaults()

  --  codec = QtCore.QTextCodec.codecForHtml(data)
  --  unistr = codec.toUnicode(data)

  -- guess encoding
  enc = $.bytes.hasBOM(data)
  if enc=false then
    if $.bytes.isUtf8(data) then enc = "utf-8"
  end if
  if enc=false then enc = "windows-1252"
  data.position = 1
  me.pTextEdit.setText(data.readRawString(data.length, enc))

  me.currentCharFormatChanged(me.pTextEdit)
  me.setCurrentFileName(fn)

  return True
end

--------------------------------------
--
--------------------------------------
on maybeSave(me)
	
  if not me.pTextEdit.isModified() then return TRUE
  --  if me.pFileName starts ":/" then return TRUE -- internal resource

  tStandardButtons = bitOr(bitOr($.Qt.QMessageBox.Save, $.Qt.QMessageBox.Discard), $.Qt.QMessageBox.Cancel)
 	msg = "The document has been modified."&$.LF&"Do you want to save your changes?"	
	ret = $.Qt.MessageBox.warning(me.pMainWindow, me.pAppName,msg,tStandardButtons)

  if ret = $.Qt.QMessageBox.Save then return me.fileSave()
  if ret = $.Qt.QMessageBox.Cancel then return FALSE

  return TRUE
end

--------------------------------------
--
--------------------------------------
on setCurrentFileName(me, fn)
  me.pFileName = fn
  me.pTextEdit.setModified(FALSE)
  if voidP(fn) then
    shownName = "untitled.txt"
  else
    -- force native path delimiter - lingo version
    shownName = $.@(fn)
    -- force native path delimiter - qt version
    -- shownName = $.Qt.Dir.new().toNativeSeparators(fn)
  end if
  me.pMainWindow.setWindowTitle(me.pAppName&" - "&shownName&"[*]")
  me.pMainWindow.setWindowModified(FALSE)
  me.pActionUndo.setEnabled(FALSE)
  me.pActionRedo.setEnabled(FALSE)
  
  --me.pActionSave.setEnabled(FALSE)--???
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on fileNew(me)
  if me.maybeSave() then
    me.resetDefaults()
    me.setCurrentFileName()
  end if
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on fileOpen(me)
  filter = "HTML-Files (*.htm *.html);;Text-Files (*.txt);;All Files (*)"
  fn = $.Qt.FileDialog.getOpenFileName (me.pMainWindow, "Open File...", void, filter)
  if fn<>"" then me.load(fn)
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on fileSave(me)
  if voidP(me.pFileName) then return me.fileSaveAs()
  	
  od = _player.itemdelimiter
  _player.itemdelimiter = "."
	ftype = the last item of me.pFileName
	_player.itemdelimiter = od
  
  if me.pTextTypes.getPos(ftype) then
    success = me.pTextEdit.write(me.pFileName, "plaintext")
  else
		success = me.pTextEdit.write(me.pFileName)
  end if
  
  if success then me.pTextEdit.setModified(FALSE)
  return success
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on fileSaveAs(me)
  filter = "HTML-Files (*.htm *.html);;Text-Files (*.txt);;All Files (*)"
  fn = $.Qt.FileDialog.getSaveFileName (me.pMainWindow, "Save as...", void, filter)
  if fn="" then return FALSE
  
  od = _player.itemdelimiter
  _player.itemdelimiter = "."
	ftype = the last item of fn
	_player.itemdelimiter = od
   
  if ["htm", "html"].getPos(ftype)=0 AND me.pTextTypes.getPos(ftype)=0 then
    put ".htm" after fn
  end if
 
  me.setCurrentFileName(fn)
  return me.fileSave()
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on fileExportODF(me)
  filter = "ODF files (*.odt);;All Files (*)"
  fn = $.Qt.FileDialog.getSaveFileName (me.pMainWindow, "Export ODF", void, filter)
  if fn="" then return FALSE
  od = _player.itemdelimiter
  _player.itemdelimiter = "."
  if the last item of fn<>"odt" then
    put ".odt" after fn
  end if
  _player.itemdelimiter = od
  return me.pTextEdit.write(fn)
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on filePrint(me)
  dialog = $.Qt.PrintDialog.new(me.pMainWindow, me.pPrinter)
  --  if me.textEdit.textCursor().hasSelection() then
  --    dlg.addEnabledOption(QtGui.QAbstractPrintDialog.PrintSelection)
  --  end if
  dialog.setWindowTitle("Print Document")
  res = dialog.exec()
  if res = $.Qt.QDialog.Accepted then
    me.pTextEdit.print(me.pPrinter)
  end if
  dialog.destroy()
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on filePrintPreview(me)  
	dialog = $.Qt.PrintPreviewDialog.new(me.pMainWindow, me.pPrinter)
	dialog.setProvider(me.pTextEdit)
	res = dialog.exec()
	dialog.destroy()
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on filePrintPdf(me)
	filter = "PDF files (*.pdf);;All Files (*)"
  fn = $.Qt.FileDialog.getSaveFileName (me.pMainWindow, "Export PDF", void, filter)
  if fn="" then return FALSE
  me.pPrinter.setOutputFormat($.Qt.QPrinter.PdfFormat)
  me.pPrinter.setOutputFileName(fn)
  me.pTextEdit.print(me.pPrinter)
end

--------------------------------------
-- @callback both action and event callback
--------------------------------------
on fileQuit(me)
	--x=1/0
  if (me.maybeSave()) then
  	
  	--save window settings
  	-- 0=default=registry, 1=INI
  	settings = $.Qt.Settings.new("Fluxus","Editor",1)  
  	settings.setValue("mainwindow/rect", string(me.pMainWindow.geometry()))
    settings.setValue("mainwindow/state", string(me.pMainWindow.windowState()))
    
    halt()
  end if
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on textBold(me)
  format = $.Qt.TextCharFormat.new()
  if me.pActionTextBold.isChecked() then
    format.setFontWeight($.Qt.QFont.Bold)
  else
    format.setFontWeight($.Qt.QFont.Normal)
  end if
  me.mergeFormatOnWordOrSelection(format)
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on textUnderline(me)
  format = $.Qt.TextCharFormat.new()
  format.setFontUnderline(me.pActionTextUnderline.isChecked())
  me.mergeFormatOnWordOrSelection(format)
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on textItalic(me)
  format = $.Qt.TextCharFormat.new()
  format.setFontItalic(me.pActionTextItalic.isChecked())
  me.mergeFormatOnWordOrSelection(format)
end

--------------------------------------
--
--------------------------------------
on textFamily(me, widgetID, family)
  format = $.Qt.TextCharFormat.new()
  format.setFontFamily(family)
  me.mergeFormatOnWordOrSelection(format)
end

--------------------------------------
--
--------------------------------------
on textSize(me, widgetID, sizeStr)
  pointSize = float(sizeStr)
  if not floatP(pointSize) then return
  if pointSize>0 then
    format = $.Qt.TextCharFormat.new()
    format.setFontPointSize(pointSize)
    me.mergeFormatOnWordOrSelection(format)
  end if
end

--------------------------------------
-- @todo
--------------------------------------
--on textStyle(me, widgetID, styleIndex) -- 0..8
--
--  --  cursor = me.textEdit.textCursor()
--  if styleIndex>0 then
--
--    styleList = [\
--      $.Qt.QTextListFormat.ListDisc,\
--      $.Qt.QTextListFormat.ListCircle,\
--      $.Qt.QTextListFormat.ListSquare,\
--      $.Qt.QTextListFormat.ListDecimal,\
--      $.Qt.QTextListFormat.ListLowerAlpha,\
--      $.Qt.QTextListFormat.ListUpperAlpha,\
--      $.Qt.QTextListFormat.ListLowerRoman,\
--      $.Qt.QTextListFormat.ListUpperRoman\
--    ]
--
--    style = styleList.getPos(styleIndex) --  QtGui.QTextListFormat.ListDisc
--    if not style then style = $.Qt.QTextListFormat.ListDisc
--
----    cursor.beginEditBlock()
----    blockFmt = cursor.blockFormat()
----    listFmt = QtGui.QTextListFormat()
--
----    if cursor.currentList() then
----      listFmt = cursor.currentList().format()
----    else
----      listFmt.setIndent(blockFmt.indent() + 1)
----      blockFmt.setIndent(0)
----      cursor.setBlockFormat(blockFmt)
----    end if
--
----    listFmt.setStyle(style)
----    cursor.createList(listFmt)
----    cursor.endEditBlock()
--  else
----    bfmt = QtGui.QTextBlockFormat()
----    bfmt.setObjectIndex(-1)
----    cursor.mergeBlockFormat(bfmt)
--  end if
--end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on textColor(me)
	col = $.Qt.ColorDialog.getColor(me.pMainWindow, me.pTextColor)
  if ilk(col)=#color then
    format = $.Qt.TextCharFormat.new()
    format.setForeground(col)
    me.mergeFormatOnWordOrSelection(format)
    me.colorChanged(col)
  end if
end

--------------------------------------
-- @callback object callback (called by actionGroup)
--------------------------------------
on textAlign(me, groupID, action)
  case (action) of
    "&Left":
      me.pTextEdit.setAlignment(bitOr($.Qt.AlignLeft, $.Qt.AlignAbsolute))
    "C&enter":
      me.pTextEdit.setAlignment($.Qt.AlignHCenter)
    "&Right":
      me.pTextEdit.setAlignment(bitOr($.Qt.AlignRight, $.Qt.AlignAbsolute))
    "&Justify":
      me.pTextEdit.setAlignment($.Qt.AlignJustify)
  end case
end

--------------------------------------
-- @callback widget callback
--------------------------------------
on currentCharFormatChanged(me, widget)
  format = me.pTextEdit.currentCharFormat()
  font = format.getFont()
  me.fontChanged(font)
  col = format.foreground()
  me.colorChanged(col)
end

--------------------------------------
-- @callback widget callback
--------------------------------------
on copyAvailable(me, widgetID, bool)
  me.pActionCut.setEnabled(bool)
  me.pActionCopy.setEnabled(bool)
end

--------------------------------------
-- @callback widget callback
--------------------------------------
on undoAvailable(me, widgetID, bool)
  me.pActionUndo.setEnabled(bool)
end

--------------------------------------
-- @callback widget callback
--------------------------------------
on redoAvailable(me, widgetID, bool)
  me.pActionRedo.setEnabled(bool)
end

--------------------------------------
-- @callback widget callback
--------------------------------------
on cursorPositionChanged (me, widget)
  me.alignmentChanged(me.pTextEdit.alignment())
end

--------------------------------------
-- @callback widget callback
--------------------------------------
on modificationChanged(me, widgetID, bool)
  me.pActionSave.setEnabled(bool)
  me.pMainWindow.setWindowModified(bool)
end

--------------------------------------
-- @callback object callback (called by clipboard object)
--------------------------------------
on clipboardDataChanged(me)
  str = me.pClipboard.text()
  me.pActionPaste.setEnabled(str<>"")
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on helpConsole (me)
	$.import("lib/console").show()
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on helpEdit (me)
	fn = $._.scriptdict.getOne(me.script)
	me.load(fn) 
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on helpAbout (me)
  msg = "Qt's Rich Text Editor demo ported to Lingo.<br><br>"&\
  "This application is based on a small runtime engine I called 'LSW', "&\
  "it was written in a code editor (UltraEdit) without ever touching the Director IDE.<br><br>"&\
  "The idea of LSW is to write batch scripts (e.g.for daily automation tasks etc.) or "&\
  "desktop applications in lingo, in a way similar to e.g. Node.js or Python.<br><br>"&\
  "More infos on LSW <a href='http://valentin.dasdeck.com/lingo/lsw/'>here</a>."
  
  $.Qt.MessageBox.about(me.pMainWindow, "About "&me.pAppName, msg)
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on helpAboutQt (me)
  app = $.Qt.Application.new()
  app.aboutQt()
end

--------------------------------------
--
--------------------------------------
on mergeFormatOnWordOrSelection(me, format)
  --  cursor = me.textEdit.textCursor()
  --  if not cursor.hasSelection() then cursor.select(QtGui.QTextCursor.WordUnderCursor)
  --  cursor.mergeCharFormat(format)
  me.pTextEdit.mergeCurrentCharFormat(format)
end

--------------------------------------
--
--------------------------------------
on fontChanged(me, font)
  family = font.family()
  fontsize = integer(font.pointSize())
  bold = font.bold()
  italic = font.italic()
  underline = font.underline()

  n = me.pComboFont.findText(family)
  me.pComboFont.setCurrentIndex(n)

  n = me.pComboSize.findText(string(fontsize))
  me.pComboSize.setCurrentIndex(n)

  me.pActionTextBold.setChecked(bold)
  me.pActionTextItalic.setChecked(italic)
  me.pActionTextUnderline.setChecked(underline)
end

--------------------------------------
--
--------------------------------------
on colorChanged(me, col)
  me.pTextColor = col
  img = image(16,16,32)
  img.fill(img.rect, me.pTextColor)
  me.pActionColor.setIcon(img)
end

--------------------------------------
--
--------------------------------------
on alignmentChanged(me, alignment)
  repeat with a in ["Left","Right","HCenter","Justify"]
    if bitAnd(alignment, value("$.Qt.Align"&a)) then
      me["pActionAlign"&a].setChecked(TRUE)
      exit repeat
    end if
  end repeat
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on undo(me)
  me.pTextEdit.undo()
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on redo(me)
  me.pTextEdit.redo()
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on cut(me)
  me.pTextEdit.cut()
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on copy(me)
  me.pTextEdit.copy()
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on paste(me)
  me.pTextEdit.paste()
end

----------------------------------------
-- @callback dropHandler
----------------------------------------
on dropHandler(me, resList)
  if (not me.maybeSave()) then return
  if $.os="win" then
    prop = "application/x-qt-windows-mime;value="&QUOTE&"FileNameW"&QUOTE
    data = resList[prop]
    if not voidP(data) then
      data.position = 1
      fn = data.readRawString(data.length, "utf-16")
      me.load(fn)
    end if
  else
    -- mac code here
  end if
end
